@@ -2,10 +2,13 @@ package quic
 
 import (
 	"context"
+	"io"
 	"sync"
 	"time"
 
 	"github.com/lucas-clemente/quic-go"
+	"github.com/lucas-clemente/quic-go/logging"
+	"github.com/lucas-clemente/quic-go/qlog"
 
 	"github.com/xtls/xray-core/common"
 	"github.com/xtls/xray-core/common/net"
@@ -15,39 +18,39 @@ import (
 	"github.com/xtls/xray-core/transport/internet/tls"
 )
 
-type sessionContext struct {
+type connectionContext struct {
 	rawConn *sysConn
-	session quic.Session
+	conn    quic.Connection
 }
 
-var errSessionClosed = newError("session closed")
+var errConnectionClosed = newError("connection closed")
 
-func (c *sessionContext) openStream(destAddr net.Addr) (*interConn, error) {
-	if !isActive(c.session) {
-		return nil, errSessionClosed
+func (c *connectionContext) openStream(destAddr net.Addr) (*interConn, error) {
+	if !isActive(c.conn) {
+		return nil, errConnectionClosed
 	}
 
-	stream, err := c.session.OpenStream()
+	stream, err := c.conn.OpenStream()
 	if err != nil {
 		return nil, err
 	}
 
 	conn := &interConn{
 		stream: stream,
-		local:  c.session.LocalAddr(),
+		local:  c.conn.LocalAddr(),
 		remote: destAddr,
 	}
 
 	return conn, nil
 }
 
-type clientSessions struct {
-	access   sync.Mutex
-	sessions map[net.Destination][]*sessionContext
-	cleanup  *task.Periodic
+type clientConnections struct {
+	access  sync.Mutex
+	conns   map[net.Destination][]*connectionContext
+	cleanup *task.Periodic
 }
 
-func isActive(s quic.Session) bool {
+func isActive(s quic.Connection) bool {
 	select {
 	case <-s.Context().Done():
 		return false
@@ -56,81 +59,81 @@ func isActive(s quic.Session) bool {
 	}
 }
 
-func removeInactiveSessions(sessions []*sessionContext) []*sessionContext {
-	activeSessions := make([]*sessionContext, 0, len(sessions))
-	for i, s := range sessions {
-		if isActive(s.session) {
-			activeSessions = append(activeSessions, s)
+func removeInactiveConnections(conns []*connectionContext) []*connectionContext {
+	activeConnections := make([]*connectionContext, 0, len(conns))
+	for i, s := range conns {
+		if isActive(s.conn) {
+			activeConnections = append(activeConnections, s)
 			continue
 		}
 
-		newError("closing quic session at index: ", i).WriteToLog()
-		if err := s.session.CloseWithError(0, ""); err != nil {
-			newError("failed to close session").Base(err).WriteToLog()
+		newError("closing quic connection at index: ", i).WriteToLog()
+		if err := s.conn.CloseWithError(0, ""); err != nil {
+			newError("failed to close connection").Base(err).WriteToLog()
 		}
 		if err := s.rawConn.Close(); err != nil {
 			newError("failed to close raw connection").Base(err).WriteToLog()
 		}
 	}
 
-	if len(activeSessions) < len(sessions) {
-		newError("active quic session reduced from ", len(sessions), " to ", len(activeSessions)).WriteToLog()
-		return activeSessions
+	if len(activeConnections) < len(conns) {
+		newError("active quic connection reduced from ", len(conns), " to ", len(activeConnections)).WriteToLog()
+		return activeConnections
 	}
 
-	return sessions
+	return conns
 }
 
-func (s *clientSessions) cleanSessions() error {
+func (s *clientConnections) cleanConnections() error {
 	s.access.Lock()
 	defer s.access.Unlock()
 
-	if len(s.sessions) == 0 {
+	if len(s.conns) == 0 {
 		return nil
 	}
 
-	newSessionMap := make(map[net.Destination][]*sessionContext)
+	newConnMap := make(map[net.Destination][]*connectionContext)
 
-	for dest, sessions := range s.sessions {
-		sessions = removeInactiveSessions(sessions)
-		if len(sessions) > 0 {
-			newSessionMap[dest] = sessions
+	for dest, conns := range s.conns {
+		conns = removeInactiveConnections(conns)
+		if len(conns) > 0 {
+			newConnMap[dest] = conns
 		}
 	}
 
-	s.sessions = newSessionMap
+	s.conns = newConnMap
 	return nil
 }
 
-func (s *clientSessions) openConnection(ctx context.Context, destAddr net.Addr, config *Config, tlsConfig *tls.Config, sockopt *internet.SocketConfig) (stat.Connection, error) {
+func (s *clientConnections) openConnection(ctx context.Context, destAddr net.Addr, config *Config, tlsConfig *tls.Config, sockopt *internet.SocketConfig) (stat.Connection, error) {
 	s.access.Lock()
 	defer s.access.Unlock()
 
-	if s.sessions == nil {
-		s.sessions = make(map[net.Destination][]*sessionContext)
+	if s.conns == nil {
+		s.conns = make(map[net.Destination][]*connectionContext)
 	}
 
 	dest := net.DestinationFromAddr(destAddr)
 
-	var sessions []*sessionContext
-	if s, found := s.sessions[dest]; found {
-		sessions = s
+	var conns []*connectionContext
+	if s, found := s.conns[dest]; found {
+		conns = s
 	}
 
-	if len(sessions) > 0 {
-		s := sessions[len(sessions)-1]
-		if isActive(s.session) {
+	if len(conns) > 0 {
+		s := conns[len(conns)-1]
+		if isActive(s.conn) {
 			conn, err := s.openStream(destAddr)
 			if err == nil {
 				return conn, nil
 			}
 			newError("failed to openStream: ").Base(err).WriteToLog()
 		} else {
-			newError("current quic session is not active!").WriteToLog()
+			newError("current quic connection is not active!").WriteToLog()
 		}
 	}
 
-	sessions = removeInactiveSessions(sessions)
+	conns = removeInactiveConnections(conns)
 	newError("dialing quic to ", dest).WriteToLog()
 	rawConn, err := internet.DialSystem(ctx, dest, sockopt)
 	if err != nil {
@@ -140,39 +143,42 @@ func (s *clientSessions) openConnection(ctx context.Context, destAddr net.Addr,
 	quicConfig := &quic.Config{
 		ConnectionIDLength: 12,
 		KeepAlive:          false,
+		Tracer: qlog.NewTracer(func(_ logging.Perspective, connID []byte) io.WriteCloser {
+			return &QlogWriter{connID: connID}
+		}),
 	}
 
 	udpConn, _ := rawConn.(*net.UDPConn)
 	if udpConn == nil {
 		udpConn = rawConn.(*internet.PacketConnWrapper).Conn.(*net.UDPConn)
 	}
-	conn, err := wrapSysConn(udpConn, config)
+	sysConn, err := wrapSysConn(udpConn, config)
 	if err != nil {
 		rawConn.Close()
 		return nil, err
 	}
 
-	session, err := quic.DialContext(context.Background(), conn, destAddr, "", tlsConfig.GetTLSConfig(tls.WithDestination(dest)), quicConfig)
+	conn, err := quic.DialContext(context.Background(), sysConn, destAddr, "", tlsConfig.GetTLSConfig(tls.WithDestination(dest)), quicConfig)
 	if err != nil {
-		conn.Close()
+		sysConn.Close()
 		return nil, err
 	}
 
-	context := &sessionContext{
-		session: session,
-		rawConn: conn,
+	context := &connectionContext{
+		conn:    conn,
+		rawConn: sysConn,
 	}
-	s.sessions[dest] = append(sessions, context)
+	s.conns[dest] = append(conns, context)
 	return context.openStream(destAddr)
 }
 
-var client clientSessions
+var client clientConnections
 
 func init() {
-	client.sessions = make(map[net.Destination][]*sessionContext)
+	client.conns = make(map[net.Destination][]*connectionContext)
 	client.cleanup = &task.Periodic{
 		Interval: time.Minute,
-		Execute:  client.cleanSessions,
+		Execute:  client.cleanConnections,
 	}
 	common.Must(client.cleanup.Start())
 }
